from cython_gsl cimport *
cimport numpy as np

ctypedef double * double_ptr
ctypedef void * void_ptr

### define single layer operator ##################################################################################

cdef double s_layer(double x, void * params) nogil:
    cdef float pi = 0.318309886183791
    cdef double a1 = (<double_ptr> params)[0]
    cdef double a2 = (<double_ptr> params)[1]
    cdef double a3 = (<double_ptr> params)[2]
    cdef double a4 = (<double_ptr> params)[3]
    cdef double xi = (<double_ptr> params)[4]
    cdef double eta = (<double_ptr> params)[5]
    r = (a1*(1-x) + a2*x)
    z = (a3*(1-x) + a4*x)
    a = r*r+xi*xi+(z-eta)*(z-eta)
    b = 2*r*xi
    abp = a+b
    m = sqrt(2*b/abp)
    return r*gsl_sf_ellint_Kcomp(m, GSL_PREC_DOUBLE)/sqrt(abp)*pi

cdef double s_layer_s(double x, void * params) nogil:
    cdef float pi = 0.318309886183791
    cdef double a1 = (<double_ptr> params)[0]
    cdef double a2 = (<double_ptr> params)[1]
    cdef double a3 = (<double_ptr> params)[2]
    cdef double a4 = (<double_ptr> params)[3]
    cdef double xi = (<double_ptr> params)[4]
    cdef double eta = (<double_ptr> params)[5]
    r = (a1*(1-x) + a2*x)
    z = (a3*(1-x) + a4*x)
    a = r*r+xi*xi+(z-eta)*(z-eta)
    b = 2*r*xi
    abp = a+b
    m = .999999999999*sqrt(2*b/abp)
    return r*gsl_sf_ellint_Kcomp(m, GSL_PREC_DOUBLE)/sqrt(abp)*pi

### define double layer operator ##################################################################################

cdef double d_layer(double x, void * params) nogil:
    cdef float pi = 0.318309886183791
    cdef double a1 = (<double_ptr> params)[0]
    cdef double a2 = (<double_ptr> params)[1]
    cdef double a3 = (<double_ptr> params)[2]
    cdef double a4 = (<double_ptr> params)[3]
    cdef double xi = (<double_ptr> params)[4]
    cdef double eta = (<double_ptr> params)[5]
    nr = (a4-a3)
    nz = (-a2+a1)
    r = (a1*(1-x) + a2*x)
    z = (a3*(1-x) + a4*x)
    a = r*r+xi*xi+(z-eta)*(z-eta)
    b = 2*r*xi
    abp = a+b
    m = sqrt(2*b/abp)
    H1 = nr*(gsl_sf_ellint_Ecomp(m, GSL_PREC_DOUBLE) - gsl_sf_ellint_Kcomp(m, GSL_PREC_DOUBLE))/(2*r*sqrt(abp))
    H2 = (nr*(r-xi)+nz*(z-eta))*gsl_sf_ellint_Ecomp(m, GSL_PREC_DOUBLE)/((a-b)*sqrt(abp))
    return r*(H1-H2)*pi

cdef double d_layer_s(double x, void * params) nogil:
    cdef float pi = 0.318309886183791
    cdef double a1 = (<double_ptr> params)[0]
    cdef double a2 = (<double_ptr> params)[1]
    cdef double a3 = (<double_ptr> params)[2]
    cdef double a4 = (<double_ptr> params)[3]
    cdef double xi = (<double_ptr> params)[4]
    cdef double eta = (<double_ptr> params)[5]
    nr = (a4-a3)
    nz = (-a2+a1)
    r = (a1*(1-x) + a2*x)
    z = (a3*(1-x) + a4*x)
    a = r*r+xi*xi+(z-eta)*(z-eta)
    b = 2*r*xi
    abp = a+b
    m = .999999999999*sqrt(2*b/abp)
    H1 = nr*(gsl_sf_ellint_Ecomp(m, GSL_PREC_DOUBLE) - gsl_sf_ellint_Kcomp(m, GSL_PREC_DOUBLE))/(2*r*sqrt(abp))
    if x != 0.5:
        H2 = 0.
    else:
        H2 = (nr*(r-xi)+nz*(z-eta))*gsl_sf_ellint_Ecomp(m, GSL_PREC_DOUBLE)/((a-b)*sqrt(abp))
    return r*(H1-H2)*pi


###################################################################################################################
### define single layer for calculation of neumann trace on boundary ##############################################

def int_S_func(int i, int j, np.ndarray[np.int64_t, ndim=2] u, np.ndarray[double, ndim=2] v):
    cdef double result, error
    cdef gsl_integration_workspace * w
    w = gsl_integration_workspace_alloc(1000)
    cdef gsl_function F
    cdef double params[5]

    xi = .5*(v[u[j,1],0]+v[u[j,0],0])
    eta = .5*(v[u[j,1],1]+v[u[j,0],1])
    a1 = v[u[i,0],0]
    a2 = v[u[i,1],0]
    a3 = v[u[i,0],1]
    a4 = v[u[i,1],1]
    params[0] = a1
    params[1] = a2
    params[2] = a3
    params[3] = a4
    params[4] = xi
    params[5] = eta

    F.function = &s_layer
    F.params = params
    gsl_integration_qag(&F, 0, 1, 0, 1e-8, 1000,2, w, &result, &error)
    gsl_integration_workspace_free(w)
    return result


def int_S_func_s(int i, int j, np.ndarray[np.int64_t, ndim=2] u, np.ndarray[double, ndim=2] v):
    cdef double result, error
    cdef gsl_integration_workspace * w
    w = gsl_integration_workspace_alloc(1000)
    cdef gsl_function F
    cdef double params[5]

    xi = .5*(v[u[j,1],0]+v[u[j,0],0])
    eta = .5*(v[u[j,1],1]+v[u[j,0],1])
    a1 = v[u[i,0],0]
    a2 = v[u[i,1],0]
    a3 = v[u[i,0],1]
    a4 = v[u[i,1],1]
    params[0] = a1
    params[1] = a2
    params[2] = a3
    params[3] = a4
    params[4] = xi
    params[5] = eta

    F.function = &s_layer_s
    F.params = params
    gsl_integration_qagp(&F, [0,.5,1],3, 0, 1e-8, 1000, w, &result, &error)
    gsl_integration_workspace_free(w)
    return result


### define double layer for calculation of neumann trace on boundary ##############################################

def int_D_func(int i, int j, np.ndarray[np.int64_t, ndim=2] u, np.ndarray[double, ndim=2] v):
    cdef double result, error
    cdef gsl_integration_workspace * w
    w = gsl_integration_workspace_alloc(1000)
    cdef gsl_function F
    cdef double params[5]

    xi = .5*(v[u[j,1],0]+v[u[j,0],0])
    eta = .5*(v[u[j,1],1]+v[u[j,0],1])
    a1 = v[u[i,0],0]
    a2 = v[u[i,1],0]
    a3 = v[u[i,0],1]
    a4 = v[u[i,1],1]
    lenght = sqrt((a2-a1)*(a2-a1) + (a4-a3)*(a4-a3))
    params[0] = a1
    params[1] = a2
    params[2] = a3
    params[3] = a4
    params[4] = xi
    params[5] = eta

    F.function = &d_layer
    F.params = params
    gsl_integration_qag(&F, 0, 1, 0, 1e-8, 1000,2, w, &result, &error)
    gsl_integration_workspace_free(w)
    return result


def int_D_func_s(int i, int j, np.ndarray[np.int64_t, ndim=2] u, np.ndarray[double, ndim=2] v):
    cdef double result, error
    cdef gsl_integration_workspace * w
    w = gsl_integration_workspace_alloc(1000)
    cdef gsl_function F
    cdef double params[5]

    xi = .5*(v[u[j,1],0]+v[u[j,0],0])
    eta = .5*(v[u[j,1],1]+v[u[j,0],1])
    a1 = v[u[i,0],0]
    a2 = v[u[i,1],0]
    a3 = v[u[i,0],1]
    a4 = v[u[i,1],1]
    lenght = sqrt((a2-a1)*(a2-a1) + (a4-a3)*(a4-a3))
    params[0] = a1
    params[1] = a2
    params[2] = a3
    params[3] = a4
    params[4] = xi
    params[5] = eta

    F.function = &d_layer_s
    F.params = params
    gsl_integration_qagp(&F, [0,.5,1],3, 0, 1e-8, 1000, w, &result, &error)
    gsl_integration_workspace_free(w)
    return result
###################################################################################################################

### define single layer for evaluation of electrical potential ##############################################

def int_S_func_ev(int i, int j, np.ndarray[np.int64_t, ndim=2] u, np.ndarray[double, ndim=2] v, np.ndarray[double, ndim=2] wz ):
    cdef double result, error
    cdef gsl_integration_workspace * w
    w = gsl_integration_workspace_alloc(1000)
    cdef gsl_function F
    cdef double params[5]
    xi = wz[j,0]
    eta = wz[j,1]
    a1 = v[u[i,0],0]
    a2 = v[u[i,1],0]
    a3 = v[u[i,0],1]
    a4 = v[u[i,1],1]
    params[0] = a1
    params[1] = a2
    params[2] = a3
    params[3] = a4
    params[4] = xi
    params[5] = eta

    F.function = &s_layer
    F.params = params
    gsl_integration_qag(&F, 0, 1, 0, 1e-8, 1000,2, w, &result, &error)
    gsl_integration_workspace_free(w)
    return result

### define double layer for evaluation of electrical potential ##############################################

def int_D_func_ev(int i, int j, np.ndarray[np.int64_t, ndim=2] u, np.ndarray[double, ndim=2] v, np.ndarray[double, ndim=2] wz ):
    cdef double result, error
    cdef gsl_integration_workspace * w
    w = gsl_integration_workspace_alloc(1000)
    cdef gsl_function F
    cdef double params[5]
    xi = wz[j,0]
    eta = wz[j,1]
    a1 = v[u[i,0],0]
    a2 = v[u[i,1],0]
    a3 = v[u[i,0],1]
    a4 = v[u[i,1],1]
    params[0] = a1
    params[1] = a2
    params[2] = a3
    params[3] = a4
    params[4] = xi
    params[5] = eta

    F.function = &d_layer
    F.params = params
    gsl_integration_qag(&F, 0, 1, 0, 1e-8, 1000,2, w, &result, &error)
    gsl_integration_workspace_free(w)
    return result


#############################################################################################################
### define single layer operator for electric field dr,dz ###################################################

cdef double s_layer_dr(double x, void * params) nogil:
    cdef float pi = 0.318309886183791
    cdef double a1 = (<double_ptr> params)[0]
    cdef double a2 = (<double_ptr> params)[1]
    cdef double a3 = (<double_ptr> params)[2]
    cdef double a4 = (<double_ptr> params)[3]
    cdef double xi = (<double_ptr> params)[4]
    cdef double eta = (<double_ptr> params)[5]
    r = (a1*(1-x) + a2*x)
    z = (a3*(1-x) + a4*x)
    dz = z-eta
    a = r*r+xi*xi+(z-eta)*(z-eta)
    b = 2*r*xi
    abp = a+b
    m = sqrt(2*b/abp)
    return r*(( (r*r-xi*xi+dz*dz)/(a-b)*gsl_sf_ellint_Ecomp(m, GSL_PREC_DOUBLE) - gsl_sf_ellint_Kcomp(m, GSL_PREC_DOUBLE) )*pi/(2*xi*sqrt(abp)))


cdef double s_layer_dz(double x, void * params) nogil:
    cdef float pi = 0.318309886183791
    cdef double a1 = (<double_ptr> params)[0]
    cdef double a2 = (<double_ptr> params)[1]
    cdef double a3 = (<double_ptr> params)[2]
    cdef double a4 = (<double_ptr> params)[3]
    cdef double xi = (<double_ptr> params)[4]
    cdef double eta = (<double_ptr> params)[5]
    r = (a1*(1-x) + a2*x)
    z = (a3*(1-x) + a4*x)
    dz = z-eta
    a = r*r+xi*xi+(z-eta)*(z-eta)
    b = 2*r*xi
    abp = a+b
    m = sqrt(2*b/abp)
    return pi*r*(dz*gsl_sf_ellint_Ecomp(m, GSL_PREC_DOUBLE)/((a-b)*sqrt(abp)))


### define single layer for evaluation of electrical field dr,dz ############################################

def int_S_func_ev_dr(int i, int j, np.ndarray[np.int64_t, ndim=2] u, np.ndarray[double, ndim=2] v, np.ndarray[double, ndim=2] wz ):
    cdef double result, error
    cdef gsl_integration_workspace * w
    w = gsl_integration_workspace_alloc(1000)
    cdef gsl_function F
    cdef double params[5]
    xi = wz[j,0]
    eta = wz[j,1]
    a1 = v[u[i,0],0]
    a2 = v[u[i,1],0]
    a3 = v[u[i,0],1]
    a4 = v[u[i,1],1]
    params[0] = a1
    params[1] = a2
    params[2] = a3
    params[3] = a4
    params[4] = xi
    params[5] = eta

    F.function = &s_layer_dr
    F.params = params
    gsl_integration_qag(&F, 0, 1, 0, 1e-8, 1000,2, w, &result, &error)
    gsl_integration_workspace_free(w)
    return result


def int_S_func_ev_dz(int i, int j, np.ndarray[np.int64_t, ndim=2] u, np.ndarray[double, ndim=2] v, np.ndarray[double, ndim=2] wz ):
    cdef double result, error
    cdef gsl_integration_workspace * w
    w = gsl_integration_workspace_alloc(1000)
    cdef gsl_function F
    cdef double params[5]
    xi = wz[j,0]
    eta = wz[j,1]
    a1 = v[u[i,0],0]
    a2 = v[u[i,1],0]
    a3 = v[u[i,0],1]
    a4 = v[u[i,1],1]
    params[0] = a1
    params[1] = a2
    params[2] = a3
    params[3] = a4
    params[4] = xi
    params[5] = eta

    F.function = &s_layer_dz
    F.params = params
    gsl_integration_qag(&F, 0, 1, 0, 1e-8, 1000,2, w, &result, &error)
    gsl_integration_workspace_free(w)
    return result


################################################################################################
### define double layer operator for electric field dr,dz ######################################

cdef double d_layer_dr(double x, void * params) nogil:
    cdef float pi = 0.318309886183791
    cdef double a1 = (<double_ptr> params)[0]
    cdef double a2 = (<double_ptr> params)[1]
    cdef double a3 = (<double_ptr> params)[2]
    cdef double a4 = (<double_ptr> params)[3]
    cdef double xi = (<double_ptr> params)[4]
    cdef double eta = (<double_ptr> params)[5]
    nr = (a4-a3)
    nz = (-a2+a1)
    r = (a1*(1-x) + a2*x)
    z = (a3*(1-x) + a4*x)
    dr = r-xi
    dz = z-eta
    a = r*r+xi*xi+(z-eta)*(z-eta)
    b = 2*r*xi
    abp = a+b
    m = sqrt(2*b/abp)
    H1 = nr*(gsl_sf_ellint_Ecomp(m, GSL_PREC_DOUBLE) - gsl_sf_ellint_Kcomp(m, GSL_PREC_DOUBLE))/(2*r*sqrt(a+b))
    H2 = (nr*(r-xi)+nz*(z-eta))*gsl_sf_ellint_Ecomp(m, GSL_PREC_DOUBLE)/((a-b)*sqrt(a+b))
    return r*(H1-H2)*pi



cdef double d_layer_dz(double x, void * params) nogil:
    cdef float pi = 0.318309886183791
    cdef double a1 = (<double_ptr> params)[0]
    cdef double a2 = (<double_ptr> params)[1]
    cdef double a3 = (<double_ptr> params)[2]
    cdef double a4 = (<double_ptr> params)[3]
    cdef double xi = (<double_ptr> params)[4]
    cdef double eta = (<double_ptr> params)[5]
    nr = (a4-a3)
    nz = (-a2+a1)
    r = (a1*(1-x) + a2*x)
    z = (a3*(1-x) + a4*x)
    dr = r-xi
    dz = z-eta
    a = r*r+xi*xi+(z-eta)*(z-eta)
    b = 2*r*xi
    abp = a+b
    m = sqrt(2*b/abp)
    NR = nr*(r-xi)+nz*(z-eta)
    H1 = ((abp*abp*(2*nz*r+nr*dz)-4*abp*r*(2*nz*r*xi + (2*NR + 3*nr*xi)*dz) + 16*r*r*xi*(NR + 2*nr*xi)*dz)*gsl_sf_ellint_Ecomp(m, GSL_PREC_DOUBLE))/pow(abp-4*r*xi,2)
    H2 = ((abp *nr - 2* r* (NR + 2 *nr *xi)) *dz *gsl_sf_ellint_Kcomp(m, GSL_PREC_DOUBLE))/(-abp+4*r*xi)
    return (H1+H2)/(2*pow(abp,1.5))

##############################################################################################
### define double layer for evaluation of electrical field dx,dy #############################

def int_D_func_ev_dr(int i, int j, np.ndarray[np.int64_t, ndim=2] u, np.ndarray[double, ndim=2] v, np.ndarray[double, ndim=2] wz ):
    cdef double result, error
    cdef gsl_integration_workspace * w
    w = gsl_integration_workspace_alloc(1000)
    cdef gsl_function F
    cdef double params[5]
    xi = wz[j,0]
    eta = wz[j,1]
    a1 = v[u[i,0],0]
    a2 = v[u[i,1],0]
    a3 = v[u[i,0],1]
    a4 = v[u[i,1],1]
    lenght = sqrt((a2-a1)*(a2-a1) + (a4-a3)*(a4-a3))
    params[0] = a1
    params[1] = a2
    params[2] = a3
    params[3] = a4
    params[4] = xi
    params[5] = eta

    F.function = &d_layer_dr
    F.params = params
    gsl_integration_qag(&F, 0, 1, 0, 1e-8, 1000,2, w, &result, &error)
    gsl_integration_workspace_free(w)
    return result


def int_D_func_ev_dz(int i, int j, np.ndarray[np.int64_t, ndim=2] u, np.ndarray[double, ndim=2] v, np.ndarray[double, ndim=2] wz ):
    cdef double result, error
    cdef gsl_integration_workspace * w
    w = gsl_integration_workspace_alloc(1000)
    cdef gsl_function F
    cdef double params[5]
    xi = wz[j,0]
    eta = wz[j,1]
    a1 = v[u[i,0],0]
    a2 = v[u[i,1],0]
    a3 = v[u[i,0],1]
    a4 = v[u[i,1],1]
    lenght = sqrt((a2-a1)*(a2-a1) + (a4-a3)*(a4-a3))
    params[0] = a1
    params[1] = a2
    params[2] = a3
    params[3] = a4
    params[4] = xi
    params[5] = eta

    F.function = &d_layer_dz
    F.params = params
    gsl_integration_qag(&F, 0, 1, 0, 1e-8, 1000,2, w, &result, &error)
    gsl_integration_workspace_free(w)
    return result
