from cython_gsl cimport *
cimport numpy as np

ctypedef double * double_ptr
ctypedef void * void_ptr

### define single layer operator ##################################################################################

cdef double s_layer(double x, void * params) nogil:
    cdef float pi = 0.318309886183791
    cdef double a1 = (<double_ptr> params)[0]
    cdef double a2 = (<double_ptr> params)[1]
    cdef double a3 = (<double_ptr> params)[2]
    cdef double a4 = (<double_ptr> params)[3]
    cdef double xi = (<double_ptr> params)[4]
    cdef double eta = (<double_ptr> params)[5]
    cdef double lenght = (<double_ptr> params)[6]
    xt = (a1*(1-x) + a2*x)
    yt = (a3*(1-x) + a4*x)
    return pi*log((xt-xi)*(xt-xi) + (yt-eta)*(yt-eta))*lenght


### define double layer operator ##################################################################################

cdef double d_layer(double x, void * params) nogil:
    cdef float pi = 0.318309886183791
    cdef double a1 = (<double_ptr> params)[0]
    cdef double a2 = (<double_ptr> params)[1]
    cdef double a3 = (<double_ptr> params)[2]
    cdef double a4 = (<double_ptr> params)[3]
    cdef double xi = (<double_ptr> params)[4]
    cdef double eta = (<double_ptr> params)[5]
    cdef double lenght = (<double_ptr> params)[6]
    nx = (a4-a3)/lenght
    ny = (-a2+a1)/lenght
    xt = (a1*(1-x) + a2*x)
    yt = (a3*(1-x) + a4*x)
    return pi*(-eta*ny + nx*(xt-xi) + ny*yt)/(eta*eta + xt*xt -2*xt*xi + xi*xi - 2*eta*yt + yt*yt)

###################################################################################################################
### define single layer for calculation of neumann trace on boundary ##############################################

def int_S_func(int i, int j, np.ndarray[np.int64_t, ndim=2] u, np.ndarray[double, ndim=2] v):
    cdef double result, error
    cdef gsl_integration_workspace * w
    w = gsl_integration_workspace_alloc(1000)
    cdef gsl_function F
    cdef double params[6]

    xi = .5*(v[u[j,1],0]+v[u[j,0],0])
    eta = .5*(v[u[j,1],1]+v[u[j,0],1])
    a1 = v[u[i,0],0]
    a2 = v[u[i,1],0]
    a3 = v[u[i,0],1]
    a4 = v[u[i,1],1]
    lenght = sqrt((a2-a1)*(a2-a1) + (a4-a3)*(a4-a3))
    params[0] = a1
    params[1] = a2
    params[2] = a3
    params[3] = a4
    params[4] = xi
    params[5] = eta
    params[6] = lenght

    F.function = &s_layer
    F.params = params
    gsl_integration_qag(&F, 0, 1, 0, 1e-8, 1000,2, w, &result, &error)
    gsl_integration_workspace_free(w)
    return result

def int_S_func_s(int i, int j, np.ndarray[np.int64_t, ndim=2] u, np.ndarray[double, ndim=2] v):
    cdef double result, error
    cdef gsl_integration_workspace * w
    w = gsl_integration_workspace_alloc(1000)
    cdef gsl_function F
    cdef double params[6]
    xi = .5*(v[u[j,1],0]+v[u[j,0],0])
    eta = .5*(v[u[j,1],1]+v[u[j,0],1])
    a1 = v[u[i,0],0]
    a2 = v[u[i,1],0]
    a3 = v[u[i,0],1]
    a4 = v[u[i,1],1]
    lenght = sqrt((a2-a1)*(a2-a1) + (a4-a3)*(a4-a3))
    params[0] = a1
    params[1] = a2
    params[2] = a3
    params[3] = a4
    params[4] = xi
    params[5] = eta
    params[6] = lenght

    F.function = &s_layer
    F.params = params
    gsl_integration_qagp(&F, [0,.5,1],3, 0, 1e-8, 1000, w, &result, &error)
    gsl_integration_workspace_free(w)
    return result

### define single layer for evaluation of electrical potential ##############################################

def int_S_func_ev(int i, int j, np.ndarray[np.int64_t, ndim=2] u, np.ndarray[double, ndim=2] v, np.ndarray[double, ndim=2] wz ):
    cdef double result, error
    cdef gsl_integration_workspace * w
    w = gsl_integration_workspace_alloc(1000)
    cdef gsl_function F
    cdef double params[6]
    xi = wz[j,0]
    eta = wz[j,1]
    a1 = v[u[i,0],0]
    a2 = v[u[i,1],0]
    a3 = v[u[i,0],1]
    a4 = v[u[i,1],1]
    lenght = sqrt((a2-a1)*(a2-a1) + (a4-a3)*(a4-a3))
    params[0] = a1
    params[1] = a2
    params[2] = a3
    params[3] = a4
    params[4] = xi
    params[5] = eta
    params[6] = lenght

    F.function = &s_layer
    F.params = params
    gsl_integration_qag(&F, 0, 1, 0, 1e-8, 1000,2, w, &result, &error)
    gsl_integration_workspace_free(w)
    return result


###################################################################################################################
### define double layer for calculation of neumann trace on boundary ##############################################

def int_D_func(int i, int j, np.ndarray[np.int64_t, ndim=2] u, np.ndarray[double, ndim=2] v):
    cdef double result, error
    cdef gsl_integration_workspace * w
    w = gsl_integration_workspace_alloc(1000)
    cdef gsl_function F
    cdef double params[6]
    xi = .5*(v[u[j,1],0]+v[u[j,0],0])
    eta = .5*(v[u[j,1],1]+v[u[j,0],1])
    a1 = v[u[i,0],0]
    a2 = v[u[i,1],0]
    a3 = v[u[i,0],1]
    a4 = v[u[i,1],1]
    lenght = sqrt((a2-a1)*(a2-a1) + (a4-a3)*(a4-a3))
    params[0] = a1
    params[1] = a2
    params[2] = a3
    params[3] = a4
    params[4] = xi
    params[5] = eta
    params[6] = lenght

    F.function = &d_layer
    F.params = params
    gsl_integration_qag(&F, 0, 1, 0, 1e-8, 1000,2, w, &result, &error)
    gsl_integration_workspace_free(w)
    return result

def int_D_func_s(int i, int j, np.ndarray[np.int64_t, ndim=2] u, np.ndarray[double, ndim=2] v):
    cdef double result, error
    cdef gsl_integration_workspace * w
    w = gsl_integration_workspace_alloc(1000)
    cdef gsl_function F
    cdef double params[6]
    xi = .5*(v[u[j,1],0]+v[u[j,0],0])
    eta = .5*(v[u[j,1],1]+v[u[j,0],1])
    a1 = v[u[i,0],0]
    a2 = v[u[i,1],0]
    a3 = v[u[i,0],1]
    a4 = v[u[i,1],1]
    lenght = sqrt((a2-a1)*(a2-a1) + (a4-a3)*(a4-a3))
    params[0] = a1
    params[1] = a2
    params[2] = a3
    params[3] = a4
    params[4] = xi
    params[5] = eta
    params[6] = lenght

    F.function = &d_layer
    F.params = params
    gsl_integration_qagp(&F, [0,.5,1],3, 0, 1e-8, 1000, w, &result, &error)
    gsl_integration_workspace_free(w)
    return result


### define double layer for evaluation of electrical potential ##############################################

def int_D_func_ev(int i, int j, np.ndarray[np.int64_t, ndim=2] u, np.ndarray[double, ndim=2] v, np.ndarray[double, ndim=2] wz ):
    cdef double result, error
    cdef gsl_integration_workspace * w
    w = gsl_integration_workspace_alloc(1000)
    cdef gsl_function F
    cdef double params[6]
    xi = wz[j,0]
    eta = wz[j,1]
    a1 = v[u[i,0],0]
    a2 = v[u[i,1],0]
    a3 = v[u[i,0],1]
    a4 = v[u[i,1],1]
    lenght = sqrt((a2-a1)*(a2-a1) + (a4-a3)*(a4-a3))
    params[0] = a1
    params[1] = a2
    params[2] = a3
    params[3] = a4
    params[4] = xi
    params[5] = eta
    params[6] = lenght

    F.function = &d_layer
    F.params = params
    gsl_integration_qag(&F, 0, 1, 0, 1e-8, 1000,2, w, &result, &error)
    gsl_integration_workspace_free(w)
    return result

#############################################################################################################
### define single layer operator for electric field dx,dy ###################################################

cdef double s_layer_dx(double x, void * params) nogil:
    cdef float pi = 0.318309886183791
    cdef double a1 = (<double_ptr> params)[0]
    cdef double a2 = (<double_ptr> params)[1]
    cdef double a3 = (<double_ptr> params)[2]
    cdef double a4 = (<double_ptr> params)[3]
    cdef double xi = (<double_ptr> params)[4]
    cdef double eta = (<double_ptr> params)[5]
    cdef double lenght = (<double_ptr> params)[6]
    xt = (a1*(1-x) + a2*x)
    yt = (a3*(1-x) + a4*x)
    return -pi*(xt-xi)/((xt-xi)*(xt-xi) + (-eta+yt)*(-eta+yt))*lenght

cdef double s_layer_dy(double x, void * params) nogil:
    cdef float pi = 0.318309886183791
    cdef double a1 = (<double_ptr> params)[0]
    cdef double a2 = (<double_ptr> params)[1]
    cdef double a3 = (<double_ptr> params)[2]
    cdef double a4 = (<double_ptr> params)[3]
    cdef double xi = (<double_ptr> params)[4]
    cdef double eta = (<double_ptr> params)[5]
    cdef double lenght = (<double_ptr> params)[6]
    xt = (a1*(1-x) + a2*x)
    yt = (a3*(1-x) + a4*x)
    return -pi*(-eta+yt)/((xt-xi)*(xt-xi) + (-eta+yt)*(-eta+yt))*lenght


### define single layer for evaluation of electrical field dx,dy ############################################

def int_S_func_ev_dx(int i, int j, np.ndarray[np.int64_t, ndim=2] u, np.ndarray[double, ndim=2] v, np.ndarray[double, ndim=2] wz ):
    cdef double result, error
    cdef gsl_integration_workspace * w
    w = gsl_integration_workspace_alloc(1000)
    cdef gsl_function F
    cdef double params[6]
    xi = wz[j,0]
    eta = wz[j,1]
    a1 = v[u[i,0],0]
    a2 = v[u[i,1],0]
    a3 = v[u[i,0],1]
    a4 = v[u[i,1],1]
    lenght = sqrt((a2-a1)*(a2-a1) + (a4-a3)*(a4-a3))
    params[0] = a1
    params[1] = a2
    params[2] = a3
    params[3] = a4
    params[4] = xi
    params[5] = eta
    params[6] = lenght

    F.function = &s_layer_dx
    F.params = params
    gsl_integration_qag(&F, 0, 1, 0, 1e-8, 1000,2, w, &result, &error)
    gsl_integration_workspace_free(w)
    return result


def int_S_func_ev_dy(int i, int j, np.ndarray[np.int64_t, ndim=2] u, np.ndarray[double, ndim=2] v, np.ndarray[double, ndim=2] wz ):
    cdef double result, error
    cdef gsl_integration_workspace * w
    w = gsl_integration_workspace_alloc(1000)
    cdef gsl_function F
    cdef double params[6]
    xi = wz[j,0]
    eta = wz[j,1]
    a1 = v[u[i,0],0]
    a2 = v[u[i,1],0]
    a3 = v[u[i,0],1]
    a4 = v[u[i,1],1]
    lenght = sqrt((a2-a1)*(a2-a1) + (a4-a3)*(a4-a3))
    params[0] = a1
    params[1] = a2
    params[2] = a3
    params[3] = a4
    params[4] = xi
    params[5] = eta
    params[6] = lenght

    F.function = &s_layer_dy
    F.params = params
    gsl_integration_qag(&F, 0, 1, 0, 1e-8, 1000,2, w, &result, &error)
    gsl_integration_workspace_free(w)
    return result

################################################################################################
### define double layer operator for electric field dx,dy ######################################

cdef double d_layer_dx(double x, void * params) nogil:
    cdef float pi = 0.318309886183791
    cdef double a1 = (<double_ptr> params)[0]
    cdef double a2 = (<double_ptr> params)[1]
    cdef double a3 = (<double_ptr> params)[2]
    cdef double a4 = (<double_ptr> params)[3]
    cdef double xi = (<double_ptr> params)[4]
    cdef double eta = (<double_ptr> params)[5]
    cdef double lenght = (<double_ptr> params)[6]
    nx = (a4-a3)/lenght
    ny = (-a2+a1)/lenght
    xt = (a1*(1-x) + a2*x)
    yt = (a3*(1-x) + a4*x)
    return pi*( -eta*eta*nx +2*ny*(xt-xi)*yt + 2*eta*(ny*(-xt+xi) + nx*yt) + nx*(xt*xt - 2*xt*xi + xi*xi - yt*yt))/((eta*eta +xt*xt - 2*xt*xi - 2*eta*yt + yt*yt)*(eta*eta +xt*xt - 2*xt*xi - 2*eta*yt + yt*yt))


cdef double d_layer_dy(double x, void * params) nogil:
    cdef float pi = 0.318309886183791
    cdef double a1 = (<double_ptr> params)[0]
    cdef double a2 = (<double_ptr> params)[1]
    cdef double a3 = (<double_ptr> params)[2]
    cdef double a4 = (<double_ptr> params)[3]
    cdef double xi = (<double_ptr> params)[4]
    cdef double eta = (<double_ptr> params)[5]
    cdef double lenght = (<double_ptr> params)[6]
    nx = (a4-a3)/lenght
    ny = (-a2+a1)/lenght
    xt = (a1*(1-x) + a2*x)
    yt = (a3*(1-x) + a4*x)
    return pi*( eta*eta*ny +2*nx*(xt-xi)*yt + 2*eta*(nx*(xt-xi) + ny*yt) - ny*(xt*xt - 2*xt*xi + xi*xi - yt*yt))/((eta*eta +xt*xt - 2*xt*xi - 2*eta*yt + yt*yt)*(eta*eta +xt*xt - 2*xt*xi - 2*eta*yt + yt*yt))

##############################################################################################
### define double layer for evaluation of electrical field dx,dy #############################

def int_D_func_ev_dx(int i, int j, np.ndarray[np.int64_t, ndim=2] u, np.ndarray[double, ndim=2] v, np.ndarray[double, ndim=2] wz ):
    cdef double result, error
    cdef gsl_integration_workspace * w
    w = gsl_integration_workspace_alloc(1000)
    cdef gsl_function F
    cdef double params[6]
    xi = wz[j,0]
    eta = wz[j,1]
    a1 = v[u[i,0],0]
    a2 = v[u[i,1],0]
    a3 = v[u[i,0],1]
    a4 = v[u[i,1],1]
    lenght = sqrt((a2-a1)*(a2-a1) + (a4-a3)*(a4-a3))
    params[0] = a1
    params[1] = a2
    params[2] = a3
    params[3] = a4
    params[4] = xi
    params[5] = eta
    params[6] = lenght

    F.function = &d_layer_dx
    F.params = params
    gsl_integration_qag(&F, 0, 1, 0, 1e-8, 1000,2, w, &result, &error)
    gsl_integration_workspace_free(w)
    return result


def int_D_func_ev_dy(int i, int j, np.ndarray[np.int64_t, ndim=2] u, np.ndarray[double, ndim=2] v, np.ndarray[double, ndim=2] wz ):
    cdef double result, error
    cdef gsl_integration_workspace * w
    w = gsl_integration_workspace_alloc(1000)
    cdef gsl_function F
    cdef double params[6]
    xi = wz[j,0]
    eta = wz[j,1]
    a1 = v[u[i,0],0]
    a2 = v[u[i,1],0]
    a3 = v[u[i,0],1]
    a4 = v[u[i,1],1]
    lenght = sqrt((a2-a1)*(a2-a1) + (a4-a3)*(a4-a3))
    params[0] = a1
    params[1] = a2
    params[2] = a3
    params[3] = a4
    params[4] = xi
    params[5] = eta
    params[6] = lenght

    F.function = &d_layer_dy
    F.params = params
    gsl_integration_qag(&F, 0, 1, 0, 1e-8, 1000,2, w, &result, &error)
    gsl_integration_workspace_free(w)
    return result
